/*
 * Indie Game Jame 2 - example game "framework"
 * 2004/3/15 checker@d6.com
 */
#include "shared/core/first.h"

// windows crap
#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <windowsx.h>

// system
#include <assert.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include <mmsystem.h>

// gl stuff
#include <gl/gl.h>
#include "glutils.h"

// physics stuff
#include "shared/physics/2d/2d_body_hull.h"
#include "shared/physics/2d/2d_body_list.h"
#include "shared/physics/2d/2d_body_box.h"
#include "shared/physics/2d/2d_body_sphere.h"
#include "shared/physics/2d/rigid/2d_rigid_sim.h"
#include "shared/physics/2d/rigid/2d_rigid_drag_force.h"
#include "shared/physics/2d/rigid/2d_rigid_gravity_force.h"
#include "shared/physics/2d/rigid/2d_rigid_spring_force.h"
#include "shared/physics/2d/rigid/2d_rigid_control_force.h"
#include "shared/input/joystick.h"
using namespace ajb;
using namespace Physics::D2;
using namespace Physics::D2::Rigid;

// misc
#include "utils.h"
#include <vector>
using namespace std;

// media lib
#include "music.h"
#include "zsound.h"
#include "bitmap.h"

#include "property_system.h"
#include "property_system_editor.h"
#include "editor.h"

// framerate mgmt
#include "../../utils/misc/framerate.h"

/********************************* globals **********************************/

// window/camera stuff
char const AppName[] = "Help the Yogi";
int WindowWidth = 1024;
int WindowHeight = 768;
bool AppQuit = false;  // stuff this to kill yourself
float FOVY = 45.0f, ZNear = 0.1f, ZFar = 1000.0f;  // whatever, we're 2D
float AspectRatio = float(WindowWidth)/float(WindowHeight); // updated in Resize
float CameraAngleRadians = 0.0, CameraX = 0.0, CameraY = 20.0, CameraZ = 40.0;
Vec2 WorldFromWindowCoords( Vec2 window );
Vec2 WindowFromWorldCoords( Vec2 world );

// win32 stuff
HWND WindowHandle;
HDC WindowDC;

// gl stuff
HGLRC WindowRC;

// functions for you to play in
void Resize( int w, int h );
void Initialize( void );
void Uninitialize( void );
void Idle( void );
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);

// system
Music music;

// physics stuff
Sim simulator;
Drag_Force    global_drag(.05f, .05f);    //a little drag helps with stability
Gravity_Force gravity(Vec2(0.0f, -1.0f));
Rigid_Body    world_body(make_Body_Null(),Vec2(0,0),0,0,0);

PROP_PARSE_THIS_FILE;
PROP_STRUCT struct global_properties
{
    float Gravity;
    
    float MouseSpringKs;
    float MouseSpringKd;
    float MouseSpringRestLength;
};
static global_properties Globals;
static property *GlobalProperties;
static property_system_editor *GlobalsEditor;

static bool MovingCamera = false;

static vector<Rigid_Body*> objs;
static vector<Binary_Force *> forces;

static Vec2 window_mouse, world_mouse;
static Joystick PS2Controller;

//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
// yoga-specific data.
#include <map>
using std::vector;
using std::string;
using std::map;
using std::pair;

struct sTexture
{
	string filename;
	GLuint texture;
};

struct sPose
{
	string name;
	string file;
	string bgtexture;
	float maxTorque;
	float minSpringTight;
	float maxSpringTight;
	Vec3 camera;
	float bodyScalar;
	float bodyBloat;
	map<string,sTexture> textures;
};

char gpPosesListFilename[] = "poses.txt";
int gCurPose = -1;
vector<sPose> gPoses;

map<long, float> gOrigAngles;
map<long, int> gDistFromAbdomen;

float gLeanScalar = 0.1f;
Vec2 gGroundLimbTorqueVec(0,0);
Constant_Force* pGroundLimbTorqueForce = 0;
float gRadianRestorePerFrame = 0.05f * (M_PI / 180.f);

bool gDisableDebugDraw = true;

float gTimeScalar = 0.22f;

bool gbFailedLevel = false;

const float gFadeDuration = 2.f;
bool gbFadingOut = false;
long gFadeStartingTimestamp = 0;
Vec3 gFadeColor(0,0,0);

const float gLevelDuration = 20.f;
float gLevelTimeLeft = 0.f;
long gLevelStartingTimestamp = 0.f;

const int kPoseNameLeft = 200;
const int kPoseNameBottom = -50;

const int kTimeBarLeft = 200;
const int kTimeBarRight = 600;
const int kTimeBarTop = -60;
const int kTimeBarBottom = -80;

const int kLeaningTextLeft = -100;
const int kLeaningTextBottom = -300;

const int kLeaningDialOuterLeft = -95;
const int kLeaningDialInnerLeft = -75;
const int kLeaningDialOuterRight = -45;
const int kLeaningDialInnerRight = -65;

const int kLeaningDialTop = -230;
const int kLeaningDialBottom = -270;

const int kMuscleTextLeft = -120;
const int kMuscleTextBottom0 = -500;
const int kMuscleTextBottom1 = -530;

const int kMuscleBarLeft = -85;
const int kMuscleBarRight = -55;
const int kMuscleBarBottom = -465;
const int kMuscleBarTop = -320;

const int kTopBoxLeft = 180;
const int kTopBoxRight = 615;
const int kTopBoxTop = -20;
const int kTopBoxBottom = -95;

const int kSideBoxLeft = -125;
const int kSideBoxRight = -10;
const int kSideBoxTop = -215;
const int kSideBoxBottom = -545;

GLuint gBGTexture = 0;
GLuint gSplashTexture = 0;
GLuint gWinTexture = 0;

float gBodyScalar = 1.f;
float gBodyBloat = 0.f;

//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////

static Binary_Force *spring = 0;


GLuint
LoadTexture(char *Filename)
{
    GLuint texture;
    
    Bitmap_Load_Result result;
    verify(LoadBitmapByExtension(Filename,&result));

    // gl stuff
    texture = gl_texture_from_bitmap(result.data, result.width, result.height,
                                     result.bytes_per_pixel);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

    // delete the bitmap
    delete [] result.data;

    return(texture);
}

Vec2 gBoxTexCoords[4] = {Vec2(0,0), Vec2(1,0), Vec2(1,1), Vec2(0,1)};

void
glBody1v(Body *body, bool Solid)
{
    if(body)
    {
		{
			Body_Hull *p = dynamic_cast<Body_Hull*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Box *p = dynamic_cast<Body_Box*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				Vec2 center = 0.25f * (p->vert(0) + p->vert(1) + p->vert(2) + p->vert(3));
				Vec2 offset0 = 0.5f * (p->vert(0) + p->vert(1)) - center;
				Vec2 offset1 = 0.5f * (p->vert(1) + p->vert(2)) - center;

				offset0 = offset0.norm();
				offset1 = offset1.norm();

				float len01 = (p->vert(1) - p->vert(0)).len();
				float len12 = (p->vert(2) - p->vert(1)).len();

				float lenscale = max(len01, len12);

				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);

					worldspace -= center;
					worldspace *= gBodyScalar;
					worldspace += center;

					float dir0 = (i == 0 || i == 1) ? 1 : -1;
					float dir1 = (i == 1 || i == 2) ? 1 : -1;
					worldspace += gBodyBloat*dir0*lenscale*offset0;
					worldspace += gBodyBloat*dir1*lenscale*offset1;

					glTexCoord2fv(gBoxTexCoords[i].m_v);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Sphere *p = dynamic_cast<Body_Sphere*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				int total_steps = 16;
				Vec2 center = p->x();
				if(Solid)
				{
					glVertex2f(center.x(), center.y());
				}
				for(int a = 0;a<total_steps;++a)
				{
					float Angle = 2.0f * 3.14159265358979323846f * ((float)a / (float)(total_steps - 1));
					float Radius = p->r();
					Vec2 arm(Radius*cos(Angle), Radius*sin(Angle));
					arm += center;
					glVertex2f(arm.x(), arm.y());
				}
				glEnd();
			}
		}

        Body_List *l = dynamic_cast<Body_List*>(body);
        if(l)
        {
            {for(int Index = 0;
                 Index < l->num_bodies();
                 ++Index)
            {
                glBody1v(&l->body(Index), Solid);
            }}
        }
    }
}

void
glForce1v(Binary_Force *Force)
{
    if(Force)
    {
        glBegin(GL_LINES);
        glVertex2f(Force->m_a.x_worldS().x(), Force->m_a.x_worldS().y());
        glVertex2f(Force->m_b.x_worldS().x(), Force->m_b.x_worldS().y());
        glEnd();
    }
}

void
glLine2f(Vec2 v0, Vec2 v1)
{
	if (gDisableDebugDraw)
	{
		return;
	}

	glBegin(GL_LINES);
    glVertex2f(v0.x(), v0.y());
    glVertex2f(v1.x(), v1.y());
    glEnd();
}

void glAngle3(Vec2 loc, Vec2 center, float angle)
{
	if (gDisableDebugDraw)
	{
		return;
	}

	Matrix22 rot;

	rot.set_rote_ccw(angle/2.f);
	Vec2 arm0 = rot * center;
	rot.set_rote_ccw(-angle/2.f);
	Vec2 arm1 = rot * center;

	glLine2f(loc, loc+arm0);
	glLine2f(loc, loc+arm1);
}

void
ClearMouseSpring(void)
{
    if(spring)
    {
        simulator.remove(*spring);
        delete spring;
        spring = 0;
    }
}

void ClearGroundLimbTorqueForce()
{
	if (pGroundLimbTorqueForce)
	{
		simulator.remove(*pGroundLimbTorqueForce);
		delete pGroundLimbTorqueForce;
		pGroundLimbTorqueForce = 0;
	}
}

Rigid_Body* GetNamedObject(const char* pName)
{
	for (int iObj=0; iObj<objs.size(); iObj++)
	{
		if (stricmp(objs[iObj]->name().c_str(), pName) == 0)
		{
			return objs[iObj];
		}
	}

	return 0;
}

void SetupGroundLimbTorqueForce()
{
	ClearGroundLimbTorqueForce();

	Rigid_Body* pGroundLimb = GetNamedObject("groundlimb");

	if (pGroundLimb)
	{
		// Figure out the right way, in world space, to torque the ground limb.
		// we want to apply a force 1 unit above its starting center (in world space)
		// to the left or right (in world space) if we're trying to lean to the left or right, respectively.
		Vec2 torqueLocW = pGroundLimb->x + Vec2(0,1);
		Vec2 torqueLocB = pGroundLimb->world_space_to_body_space(torqueLocW);

		Vec2 torqueDirW = torqueLocW + Vec2(1,0);
		Vec2 torqueDirB = pGroundLimb->world_space_to_body_space(torqueDirW);
		torqueDirB -= torqueLocB;
		gGroundLimbTorqueVec = torqueDirB;

		Handle hGroundLimbTorque(*pGroundLimb, torqueLocB);
		pGroundLimbTorqueForce = new Constant_Force(hGroundLimbTorque);
		pGroundLimbTorqueForce->m_f = Vec2(0.f,0.f);
		simulator.add(*pGroundLimbTorqueForce);
	}
}

void
ClearAll(void)
{
    ClearMouseSpring();
	ClearGroundLimbTorqueForce();
    
    for(int i = 0;i < objs.size();++i)
    {
        assert(objs[i]);
        simulator.remove(*objs[i]);
    }

    for(int i = 0;i < forces.size();++i)
    {
        assert(forces[i]);
        simulator.remove(*forces[i]);
    }

    forces.clear();
    objs.clear();
	gOrigAngles.clear();
	gDistFromAbdomen.clear();

	gbFailedLevel = false;
	gbFadingOut = false;
}

void LoadPosesListFromFile()
{
	gPoses.clear();
	gCurPose = -1;

	FILE* pFile = fopen(gpPosesListFilename, "rt");
	if (pFile)
	{
		while (!feof(pFile))
		{
			sPose pose;

			char buf[1024] = {0};
			char lineBuffer[1024] = {0};

			fgets(lineBuffer, 1024, pFile);
			sscanf(lineBuffer, "name=%[^\n]", &buf);
			pose.name = buf;
			
			fgets(lineBuffer, 1024, pFile);
			sscanf(lineBuffer, "file=%s", &buf);
			pose.file = buf;

			fgets(lineBuffer, 1024, pFile);
			sscanf(lineBuffer, "bgtexture=%s", &buf);
			pose.bgtexture = buf;

			fgets(lineBuffer, 1024, pFile);
			sscanf(lineBuffer, "maxtorque=%f", &pose.maxTorque);

			fgets(lineBuffer, 1024, pFile);
			sscanf(lineBuffer, "minmuscle=%f", &pose.minSpringTight);

			fgets(lineBuffer, 1024, pFile);
			sscanf(lineBuffer, "maxmuscle=%f", &pose.maxSpringTight);

			float camx, camy, camz;
			fgets(lineBuffer, 1024, pFile);
			sscanf(lineBuffer, "camera=%f,%f,%f", &camx, &camy, &camz);
			pose.camera = Vec3(camx, camy, camz);

			fgets(lineBuffer, 1024, pFile);
			sscanf(lineBuffer, "bodyscale=%f", &pose.bodyScalar);

			fgets(lineBuffer, 1024, pFile);
			sscanf(lineBuffer, "bodybloat=%f", &pose.bodyBloat);

			bool bFoundTex = true;
			do
			{
				char texBuf[512] = {0};
				char bodyname[128];
				char texname[128];
				fgets(texBuf, 1024, pFile);
				if (sscanf(texBuf, "texture=%[^,],%s", &bodyname, &texname) == 2)
				{
					pose.textures[bodyname].filename = texname;
					pose.textures[bodyname].texture = LoadTexture(texname);
				}
				else
				{
					bFoundTex = false;
				}

			}  while (bFoundTex);

			gPoses.push_back(pose);
		}

		fclose(pFile);
	}
}

void CalcDistancesFromAbdomenRec(Rigid_Body* pBody, int curDistance)
{
	// if this guy is non-null and we haven't already found him, set him up.
	// note that since our figures never contain limb cycles, we can do this; otherwise we'd have to say 
	// "if he was found earlier, then check if the new distance is shorter and redo everything" or generally
	// use a smarter graph algorithm.
	if (pBody && gDistFromAbdomen.find(long(pBody)) != gDistFromAbdomen.end())
	{
		gDistFromAbdomen[long(pBody)] = curDistance;

		// now for all limbs linked to this guy, do the same for them (they're one step further away, of course.)
		for (int iForce=0; iForce<forces.size(); iForce++)
		{
			Binary_Force* pForce = dynamic_cast<Binary_Force*>(forces[iForce]);
			if (pForce)
			{
				if (&pForce->m_a.rb() == pBody)
				{
					CalcDistancesFromAbdomenRec(&pForce->m_b.rb(), curDistance+1);
				}
				else if (&pForce->m_b.rb() == pBody)
				{
					CalcDistancesFromAbdomenRec(&pForce->m_a.rb(), curDistance+1);
				}
			}
		}
	}
}

void CalcDistancesFromAbdomen()
{
	// start with the abdomen.
	Rigid_Body* pAbdomen = GetNamedObject("abdomen");
	
	// now recurse.
	CalcDistancesFromAbdomenRec(pAbdomen, 0);
}

void
LoadNextPose()
{
	ClearAll();

	if (gCurPose < int(gPoses.size())-1)
	{
		int num_rb=0, num_force=0;
		Rigid_Body **rb_list;
		Binary_Force **force_list;

		loadPhysicsFile(const_cast<char*>(gPoses[gCurPose+1].file.c_str()), 0,22, 1.0/16, 1.0/16,
						&rb_list, &num_rb, &force_list, &num_force, &world_body);

		for (int i=0; i < num_rb; ++i)
		{
			objs.push_back(rb_list[i]);
			simulator.add(*rb_list[i]);
		}

		for (int i=0; i < num_force; ++i)
		{
			forces.push_back(force_list[i]);
			simulator.add(*force_list[i]);

			// if this is a joint spring, remember its original angle so if we lean the character
			// left or right we know what to reset to.
			Angle_Spring_Force* pMuscle = dynamic_cast<Angle_Spring_Force*>(forces[i]);
			if (pMuscle)
			{
				gOrigAngles[long(pMuscle)] = pMuscle->m_rest_len;
			}
		}

		gCurPose++;

		// calculate, for each limb, how far it is from the abdomen.
		CalcDistancesFromAbdomen();

		// setup the constant force we use to torque the ground limb for balance.
		SetupGroundLimbTorqueForce();

		// note the starting time.
		gLevelStartingTimestamp = timeGetTime();

		// load the bg texture for this level.
		glDeleteTextures(1, &gBGTexture);
		gBGTexture = LoadTexture((char*)gPoses[gCurPose].bgtexture.c_str());

		CameraX = gPoses[gCurPose].camera.x();
		CameraY = gPoses[gCurPose].camera.y();
		CameraZ = gPoses[gCurPose].camera.z();

		gBodyScalar = gPoses[gCurPose].bodyScalar;
		gBodyBloat = gPoses[gCurPose].bodyBloat;
	}
	else
	{
		gCurPose = gPoses.size();
	}
}

void ReloadCurrentPose()
{
	gCurPose--;
	LoadNextPose();
}

void LoadPreviousPose()
{
	if (gCurPose > 0)
	{
		gCurPose -= 2;
		LoadNextPose();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Futzes with the joint angles to attempt to lean the character to the
// left (if value is negative) or right (if positive.)  Value should be in
// range (-1,1).
void LeanPerson(float direction)
{
	// Okay, for every muscle force (angle spring force linking two
	// bodies) try to solve for the world space orientation of the joint.
	// if they want us to lean the character to the left, and the joint's
	// opening substantially to the left, tighten the angle, but if it's opening
	// substantially to the right, loosen it (to a max of 180 degrees.)
	// vice-versa if they want the character leaning to the right.

	// the first, easy thing to do is to set the force on the ground limb to try to torque it in the right way to help balance.
	if (pGroundLimbTorqueForce)
	{
		Vec2 torqueForce = (direction * gPoses[gCurPose].maxTorque) * gGroundLimbTorqueVec;
		pGroundLimbTorqueForce->m_f = torqueForce;
		
		bool bDebugDrawOff = gDisableDebugDraw;
		//gDisableDebugDraw = false;
		Vec2 worldLoc = pGroundLimbTorqueForce->m_a.safe_x_worldS();

		Handle hDummy(pGroundLimbTorqueForce->m_a.rb(), pGroundLimbTorqueForce->m_a.x_bodyS() + gGroundLimbTorqueVec);
		Vec2 targLoc = hDummy.safe_x_worldS();

		targLoc -= worldLoc;
		targLoc *= 5.f * direction;

		glColor3f(1.f,0.f,1.f);
		glLineWidth(2);
		glLine2f(worldLoc, worldLoc+targLoc);

		gDisableDebugDraw = bDebugDrawOff;
	}

	for (int iForce=0; iForce<forces.size(); iForce++)
	{
		Angle_Spring_Force* pMuscle = dynamic_cast<Angle_Spring_Force*>(forces[iForce]);
		if (pMuscle)
		{
			// Rest this guy, first, to his original orientation.  Otherwise joints can get set to different values
			// and then move to where they never get changed back because they're not "useful" for leaning.
			if (gOrigAngles.find(long(pMuscle)) != gOrigAngles.end())
			{
				// move it at however many degrees per frame they want back towards its original pose.
				float origAngle = gOrigAngles[long(pMuscle)];
                if (pMuscle->m_rest_len < origAngle)
				{
					float adjustAmt = min(gRadianRestorePerFrame, origAngle - pMuscle->m_rest_len);
					pMuscle->m_rest_len += adjustAmt;
				}
				else
				{
					float adjustAmt = min(gRadianRestorePerFrame, pMuscle->m_rest_len - origAngle);
					pMuscle->m_rest_len -= adjustAmt;
				}
			}			

			// get the two limbs this muscle connects.
			Handle ha = pMuscle->m_a;
			Handle hb = pMuscle->m_b;

			Rigid_Body* pObjA = &ha.rb();
			Rigid_Body* pObjB = &hb.rb();

			if (pObjA && pObjB)
			{
				// There's also a point-to-point force holding these together; find it.
				for (int iP2PForce=0; iP2PForce<forces.size(); iP2PForce++)
				{
					Pt2Pt_Constraint* pHinge = dynamic_cast<Pt2Pt_Constraint*>(forces[iP2PForce]);
					if (pHinge)
					{
						// see if this constraint is the one connecting these two particular limbs.
						Rigid_Body* pa2 = &pHinge->m_a.rb();
						Rigid_Body* pb2 = &pHinge->m_b.rb();

						// maybe i drew them in the opposite direction, so check both ways.
						if ((pa2 == pObjA && pb2 == pObjB) || (pa2 == pObjB && pb2 == pObjA))
						{
							// if the constraints were drawn in opposite orders make sure to flip these handles
							// so the handles for the two match.
							Handle hHingeA = pa2 == pObjA ? pHinge->m_a : pHinge->m_b;
							Handle hHingeB = pa2 == pObjA ? pHinge->m_b : pHinge->m_a;

							// okay, so we've got the constraint.
							// so now p is the world-space location of the joint constraint.
							Vec2 p = hHingeA.safe_x_worldS();

							// get the centroids of the objects locally here just for clarity.
							Vec2 ca = pObjA->x;
							Vec2 cb = pObjB->x;

							// now the vectors p->a && p->b form the current angle of the joint.
							Vec2 pa = ca - p;
							Vec2 pb = cb - p;

							glLineWidth(1);
							glColor3f(0.f, 0.f, 0.75f);
							glLine2f(p, ca);
							glLine2f(p, cb);
							glLineWidth(1);

							// now the half-angle vector between those two points the way the joint
							// is "opening" -- e.g. if the joint is an elbow, h is going to be a vector
							// pointing straight out of the eye of the elbow.
							Vec2 h = pb + pa;

							// if the joint is open to exactly 180 degrees we don't really know what to do with
							// it so only deal with joints not open all the way
							if (h.len2() > 0.0001)
							{
								h *= 1.f/h.len();

								// okay, now if fabs(x) > cos(sidewaysAngle) then the joint is pointing sideways enough
								// that it's worth mucking with to try to balance.
								const float cosSidewaysAngle = cosf(45.f);

								if (fabs(h.x()) > cosSidewaysAngle)
								{
									glColor3f(0.4f,0.4f,0.4f);
									glLineWidth(2);
								}
								else
								{
									glColor3f(0.4f,0.4f,0.4f);
									glLineWidth(2);
								}

								glLine2f(p, p+(5.f*h));
								glLineWidth(1);

								bool bJointOpensSideways = fabs(h.x()) > cosSidewaysAngle;

								// if the joint doesn't open out we'll need to know which of the two bodies is the parent (i.e. closer to the
								// abdomen) and which is the child.
								Vec2 parentLoc;
								Vec2 childLoc;
								Vec2 abdomenLoc;
								if (gDistFromAbdomen.find(long(pObjA)) != gDistFromAbdomen.end() && gDistFromAbdomen.find(long(pObjB)) != gDistFromAbdomen.end())
								{
									if (gDistFromAbdomen[long(pObjA)] < gDistFromAbdomen[long(pObjB)])
									{
										parentLoc = pObjA->x;
										childLoc = pObjB->x;
									}
									else
									{
										parentLoc = pObjB->x;
										childLoc = pObjA->x;
									}
								}

								Rigid_Body* pAbdomen = GetNamedObject("abdomen");
								if (pAbdomen)
								{
									abdomenLoc = pAbdomen->x;
								}

								// @NOTE BHS we can solve for the right "fully open" angle as a combo of:
								// (a) does joint aim for more OR less than 180 degrees, and (b) were the rectangles
								// both vertical / horizontal on construction OR was one vertical and one horizontal?
								bool bRectsBeganPerpendicular = false;
								Body_Box* pBoxA = dynamic_cast<Body_Box*>(pObjA->m_Pbody.get());
								Body_Box* pBoxB = dynamic_cast<Body_Box*>(pObjB->m_Pbody.get());

								if (pBoxA && pBoxB)
								{
									float widthA = 2.f* fabs(pBoxA->bs_vert(0).x());
									float heightA = 2.f* fabs(pBoxA->bs_vert(0).y());
									float widthB = 2.f* fabs(pBoxB->bs_vert(0).x());
									float heightB = 2.f* fabs(pBoxB->bs_vert(0).y());

									// did rectangle A begin life vertical in the editor?
									bool bAVertical = widthA < heightA;

									// did rectangle B begin life vertical in the editor?
									bool bBVertical = widthB < heightB;

									// now figure out which edge of each rectangle the constraint is placed on.
									// @NOTE BHS if it's placed near a corner, then technically accept it as on
									// either face (by saying that it is both placed vertically AND horizontally)
									// since that's probably what i meant to do when i built it in the editor.
									Vec2 normAConstraintBS = hHingeA.x_bodyS();
									normAConstraintBS.set_x(2.f*normAConstraintBS.x() / widthA);
									normAConstraintBS.set_y(2.f*normAConstraintBS.y() / heightA);
									bool bAConstrainedVert = fabs(normAConstraintBS.y()) > 0.85f;
									bool bAConstrainedHoriz = fabs(normAConstraintBS.x()) > 0.85f;

									Vec2 normBConstraintBS = hHingeB.x_bodyS();
									normBConstraintBS.set_x(2.f*normBConstraintBS.x() / widthB);
									normBConstraintBS.set_y(2.f*normBConstraintBS.y() / heightB);
									bool bBConstrainedVert = fabs(normBConstraintBS.y()) > 0.8f;
									bool bBConstrainedHoriz = fabs(normBConstraintBS.x()) > 0.8f;

									bool bAOnShortEdge = bAVertical ? bAConstrainedVert : bAConstrainedHoriz;
									bool bBOnShortEdge = bBVertical ? bBConstrainedVert : bBConstrainedHoriz;

									// okay, we only want to screw with limbs that are attached at the short end -- that is,
									// we don't want to lean the neck and other stubbier 'limbs'.
									//
									// so basically if a rectangle began life vertical, it should be constrained vertically too,
									// or vice versa.
									if (true)//bAOnShortEdge && bBOnShortEdge)
									{
										// if one was wider than tall and the other was not, they began perpendicular.
										bRectsBeganPerpendicular = (widthA > heightA) != (widthB > heightB);

										float closedAngle;

										if (bAVertical)
										{
											bool bAConstrainedBottom = normAConstraintBS.y() < 0;
											if (bAConstrainedBottom)
											{
												if (bBVertical)
												{
													bool bBConstrainedBottom = normBConstraintBS.y() < 0;

													// if both are constrained at the bottom then they're closed at rest (angle 0)
													// otherwise they're fully open at rest (closed at 180 degrees.)
													closedAngle = bBConstrainedBottom ? 0 : M_PI;
												}
												else
												{
													bool bBConstrainedLeft = normBConstraintBS.x() < 0;

													// right-hand rule: if b is constrained at left and a at bottom, then they're
													// closed when b is rotated +90 degrees, i.e. the delta angle b-a is 90.
													// otherwise they're closed at -90 degrees (if b is constrained on the other side.)
													closedAngle = bBConstrainedLeft ? M_PI*0.5f : -M_PI*0.5f;
												}
											}
											else
											{
												if (bBVertical)
												{
													bool bBConstrainedBottom = normBConstraintBS.y() < 0;

													// if both are constrained at the top then they're closed at rest (angle 0)
													// otherwise they're fully open at rest (closed at 180 degrees.)
													closedAngle = bBConstrainedBottom ? M_PI : 0;
												}
												else
												{
													bool bBConstrainedLeft = normBConstraintBS.x() < 0;

													// right-hand rule: if b is constrained at left and a at top, then they're
													// closed when b is rotated -90 degrees, i.e. the delta angle b-a is -90.
													// otherwise they're closed at 90 degrees (if b is constrained on the other side.)
													closedAngle = bBConstrainedLeft ? -M_PI*0.5f : M_PI*0.5f;
												}
											}
										}
										else
										{
											bool bAConstrainedLeft = normAConstraintBS.x() < 0;
											if (bAConstrainedLeft)
											{
												if (bBVertical)
												{
													bool bBConstrainedBottom = normBConstraintBS.y() < 0;

													// right-hand rule: if a is constrained at left and b at bottom, then they're
													// closed when b is rotated -90 degrees, i.e. the delta angle b-a is -90.
													// otherwise they're closed at 90 degrees (if b is constrained on the other side.)
													closedAngle = bBConstrainedBottom ? -M_PI*0.5f : M_PI*0.5f;
												}
												else
												{
													bool bBConstrainedLeft = normBConstraintBS.x() < 0;

													// if both are constrained at the left then they're closed at rest (angle 0)
													// otherwise they're fully open at rest (closed at 180 degrees.)
													closedAngle = bBConstrainedLeft ? 0 : M_PI;
												}
											}
											else
											{
												if (bBVertical)
												{
													bool bBConstrainedBottom = normBConstraintBS.y() < 0;

													// right-hand rule: if a is constrained at right and b at bottom, then they're
													// closed when b is rotated 90 degrees, i.e. the delta angle b-a is 90.
													// otherwise they're closed at -90 degrees (if b is constrained on the other side.)
													closedAngle = bBConstrainedBottom ? M_PI*0.5f : -M_PI*0.5f;
												}
												else
												{
													bool bBConstrainedLeft = normBConstraintBS.x() < 0;

													// if both are constrained at the right then they're closed at rest (angle 0)
													// otherwise they're fully open at rest (closed at 180 degrees.)
													closedAngle = bBConstrainedLeft ? M_PI : 0;
												}
											}
										}

										// now the fully-open angle is 180 degrees off the closed angle.  calculate
										// it and wrap to the range [0,360)
										float openAngle = closedAngle + M_PI;
										while (openAngle >= 2*M_PI)
										{
											openAngle -= 2*M_PI;
										}

										if (gOrigAngles.find(long(pMuscle)) != gOrigAngles.end())
										{
											// first, wrap the original angle d to the range [0,360].
											float rawOrigAngle = gOrigAngles[long(pMuscle)];

											float origAngle = rawOrigAngle;
											while (origAngle < 0.f)
											{
												origAngle += 2*M_PI;
											}
											while (origAngle >= 2*M_PI)
											{
												origAngle -= 2*M_PI;
											}

											// calculate whether we're trying to open or close this joint -- close if the joint is opening in the
											// direction we're trying to lean; open if the joint is opening away from the direction we're trying
											// to lean.  
											//
											// the joint is opening in the direction we want to lean if their sign is the same (i.e. the product
											// of the two is positive.)
											bool bClosingJoint = false;
											if (bJointOpensSideways)
											{
												bClosingJoint = (direction * h.x()) > 0.f;
											}
											else
											{
												// determine if the more-parent joint is closer to the center of mass than the more-child joint,
												// (here we approximate the center of mass as the centroid of the abdomen, since, hey,
												// it's probably around there)
												bool bChildFurtherFromCoM = fabs(parentLoc.x() - abdomenLoc.x()) < fabs(childLoc.x() - abdomenLoc.x());

												// also determine if the joint is to the left of the center of mass or the right.  use the point
												// constraint location (the actual location of the joint hinge) here instead of either object's
												// centroid.  if the sign of that (the direction from abdomen to joint) and the lean direction
												// are the same -- if their product is positive -- then the joint is in the direction we're trying
												// to lean.
												bool bJointInLeanDirection = ((p.x() - abdomenLoc.x()) * direction) > 0.f;

												// okay, now from all that info we can determine whether we want to open or close the joint; if
												// the joint is to the direction we want to lean and the child limb is further from the center
												// of mass, then open the joint up to push that child limb even further away (shifting the center
												// of mass to the lean direction.)  but if all of that is true but the child limb is closer to the
												// center of mass, then close it -- again, to pull the child limb away from the center of mass.
												// and flip all of that if the joint is not on the side of the body we're trying to lean towards.

												// so this is true (we want to close the joint if the child is further from the CoM) if the
												// joint is not on the side of the body we're leaning towards.
												bClosingJoint = bChildFurtherFromCoM;

												// ... so if it IS on the side we're leaning towards, flip that.
												if (bJointInLeanDirection)
												{
													bClosingJoint = !bClosingJoint;		
												}
											}
											float bendAmt = fabs(direction) * gLeanScalar;
											
											// so now we have the fully-closed angle, the original angle, and the fully-open angle.
											//
											// and we know whether we're supposed to be opening or closing the joint.
											//
											// and we have a scalar, bendAmt, in the range [0,1] for indicating how far to move from
											// the original angle to either the closed or open angle, depending on what we're trying to do.
											float extremeAngle = bClosingJoint ? closedAngle : openAngle;

											// make sure we rotate "the short way around" -- i.e. if extremeAngle is -90 degrees
											// and origAngle is 200 degrees, really we want extremeAngle to be 270 degrees instead.
											while (fabs(extremeAngle - origAngle) > M_PI)
											{
												if (extremeAngle < origAngle)
												{
													extremeAngle += 2*M_PI;
												}
												else
												{
													origAngle += 2*M_PI;
												}
											}

											float newAngle = origAngle + (bendAmt * (extremeAngle - origAngle));

											// okay, while the angle -90 degrees is "the same as" the angle 270 degrees statically,
											// if we set a joint constraint from -90 to 270 it will interpolate between the two,
											// thrashing around in a complete circle.  so it's important to try to reduce newAngle
											// to its off-by-2-pi equivalent closest to the original angle.
											while (fabs(newAngle - rawOrigAngle - 2*M_PI) < fabs(newAngle - rawOrigAngle))
											{
												newAngle -= 2*M_PI;
											}
											while (fabs(newAngle - rawOrigAngle + 2*M_PI) < fabs(newAngle - rawOrigAngle))
											{
												newAngle += 2*M_PI;
											}

											// draw the new target joint flex.
											glLineWidth(2);

											glColor3f(0.75f,0.f,0.f);
											glAngle3(p, 5.f*h, newAngle);

											glColor3f(0.f,0.5f,0.75f);
											glAngle3(p, 5.f*h, rawOrigAngle);

											glColor3f(0.f,0.f,0.f);
											glAngle3(p, 5.f*h, origAngle);

											glLineWidth(1);

											static bool bHackDisableLean = false;
											if (!bHackDisableLean)
											{
												static bool bHackDisableSideways = false;
												if (!bHackDisableSideways || !bJointOpensSideways)
												{
													pMuscle->m_rest_len = newAngle;
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}
}

/******************************** Initialize ********************************/
/* the window and graphics subsystem are set up but still hidden            */
void Initialize( void )
{
	InitializeFrameRate(60.f);

	PS2Controller.init();

    Globals.Gravity = -100.0f;

    Globals.MouseSpringKs = 128.0f;
    Globals.MouseSpringKd = 32.0f;
    Globals.MouseSpringRestLength = 0.0f;

    GlobalProperties = CloneProperty(PROP_LOCAL_STRUCT(global_properties));
    ConvertStructToProperty(PROP_LOCAL_STRUCT(global_properties), &Globals,
                            GlobalProperties);
    
    GetTime(); // throw away initial 0.0 so we're seconds-from-startup

    // gl stuff
    glClearColor(1.0f,1.0f,1.0f,0.0f);
    glEnable(GL_CULL_FACE);
    glEnable(GL_NORMALIZE);
    glEnable(GL_DEPTH_TEST);
    glShadeModel(GL_SMOOTH);

    // init our view
    Resize(WindowWidth,WindowHeight);

    // music init
    music.init();
    music.set_subdirectory("content/music/");
    music.setupnotes();
    music.loadseg("background", "simple_loop.sgt", MUSIC_TRACK_MAIN);

    // sound init
    zsoundInit(WindowHandle);

    // initialize the physics
    simulator.add(global_drag);
    simulator.add(gravity);
    simulator.add(world_body);

	//////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////
	// yoga-specific init.
	LoadPosesListFromFile();

	gSplashTexture = LoadTexture("art/splash.jpg");
	gWinTexture = LoadTexture("art/win.jpg");
}

bool IsBodyAllowedToHitFloor(Rigid_Body* pBody)
{
	const int kCanHitFloorMask = (1 << 25) | (1 << 26) | (1 << 27) | (1 << 28);
	return (pBody->m_Pbody->m_collision_plane & kCanHitFloorMask) != 0;
}

void HandleCollisions()
{
	for (Contact_List::const_iterator cit=simulator.m_contacts.begin(); cit != simulator.m_contacts.end(); cit++)
	{
		// things to do when checking collisions:
		//
		// 1.) if anything collides with the object named "floor" and the thing colliding with the floor is not a limb
		// allowed to hit the ground (which we hack by putting allowed limbs in either collision groups 24 or 25) then
		// they've lost; reload the level.
		Rigid_Body* pGameObj1 = (Rigid_Body *)(cit->m_Prb_a);
		Rigid_Body* pGameObj2 = (Rigid_Body *)(cit->m_Prb_b);

		if (pGameObj1 && pGameObj2 && !gbFailedLevel)
		{
			if (stricmp(pGameObj1->name().c_str(), "floor") == 0)
			{
				if (!IsBodyAllowedToHitFloor(pGameObj2))
				{
					gbFailedLevel = true;
					gbFadingOut = true;
					gFadeColor = Vec3(0,0,0);
					gFadeStartingTimestamp = timeGetTime();
				}
			}
			if (stricmp(pGameObj2->name().c_str(), "floor") == 0)
			{
				if (!IsBodyAllowedToHitFloor(pGameObj1))
				{
					gbFailedLevel = true;
					gbFadingOut = true;
					gFadeColor = Vec3(0,0,0);
					gFadeStartingTimestamp = timeGetTime();
				}
			}
		}
	}
}

/*********************************** Idle ***********************************/
/* called by the message loop, and by paint messages                        */
void Idle( void )
{
    ConvertPropertyToStruct(FindSubPropertyByName(LocalTemplates, "global_properties"),
                            GlobalProperties,
                            &Globals);

    gravity.m_g.set_y(Globals.Gravity);
    
	PS2Controller.sample();
	Vec2 LeftStick = PS2Controller.left_stick();
	Vec2 RightStick = PS2Controller.right_stick();

	//////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////
	// yoga-specific game logic.

	// update the time left.
	long curTime = timeGetTime();
	gLevelTimeLeft = gLevelDuration - ((curTime - gLevelStartingTimestamp)/1000.f);

	if (gCurPose != -1 && gLevelTimeLeft <= 0 && !gbFadingOut)
	{
		gbFailedLevel = false;
		gbFadingOut = true;
		gFadeColor = Vec3(1,1,1);
		gFadeStartingTimestamp = timeGetTime();
	}

	// wipe thumbsticks if we're fading out.
	if (gbFadingOut)
	{
		LeftStick = Vec2(0,0);
		RightStick = Vec2(0,0);
	}

	float curFadeCompletion = 0.f;
	if (gbFadingOut)
	{
		// calc current fade completion 0 - 1 where 1 means "done".
		curFadeCompletion = (gFadeDuration - ((curTime - gFadeStartingTimestamp)/1000.f)) / gFadeDuration;
		curFadeCompletion = 1.f - curFadeCompletion;

		// if the fade finished, what to do?
		if (curFadeCompletion >= 1.f)
		{
			if (gbFailedLevel)
			{
				ReloadCurrentPose();
			}
			else
			{
				LoadNextPose();
			}
		}
	}

	// if they've won the game and they hit X, they want to start again.
	if (gCurPose == gPoses.size() && PS2Controller.button_went_down(Joystick::BUTTON_RIGHT_D))
	{
		gCurPose = -1;
		ClearAll();
	}

	// if they're in the splash screen
	if (gCurPose == -1 && PS2Controller.button_went_down(Joystick::BUTTON_RIGHT_D))
	{
		gbFailedLevel = false;
		gbFadingOut = true;
		gFadeColor = Vec3(1,1,1);
		gFadeStartingTimestamp = timeGetTime();
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	// Adjust all spring forces according to the y axis of the right thumbstick.
	float springForce = RightStick.y();

	// move from [-1,1] to [0,1];
	springForce += 1.f;
	springForce *= 0.5f;

	// now scale from [0,1] to [gkMinSpringForce,gkMaxSpringForce]
	springForce *= gPoses[gCurPose].maxSpringTight-gPoses[gCurPose].minSpringTight;
	springForce += gPoses[gCurPose].minSpringTight;

	for (int iForce=0; iForce<forces.size(); iForce++)
	{
		Angle_Spring_Force* pMuscle = dynamic_cast<Angle_Spring_Force*>(forces[iForce]);
		if (pMuscle)
		{
			pMuscle->m_Ks = springForce;
			pMuscle->m_Kd = 300.f;
		}
	}

	/**************************** step the simulator ****************************/   

	// don't run the sim if they won, or the guy might fall over while we're fading out.
	// let him keep going and totally bite it if they lost, though, it's funny.  ha ha.
	if (!gbFadingOut || gbFailedLevel)
	{
		if (spring)                             //update
			spring->m_a = Handle(world_body, world_mouse);

		int const NumberOfSimSteps = 1;
		for(int StepCounter = 0;StepCounter < NumberOfSimSteps;++StepCounter) {
			// step as many times as you like
			simulator.step(gTimeScalar/(NumberOfSimSteps * GetFrameRate()));

			// do any game stuff you want here (like per-step collision response, etc.)
			HandleCollisions();
		}
		// now wait for the rest of the time so we're consistent on fast machines
		WaitFrameRate(GetFrameRate());
	}

    /**************************** render everything *****************************/
   
    // lots of gl stuff
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);

	// draw background.
	int vp[4];
	glGetIntegerv(GL_VIEWPORT,vp);
	int pixW = vp[2];
	int pixH = vp[3];

	if (gCurPose >= 0 && gCurPose < gPoses.size())
	{
		PushOrthoPixels();
		glColor3f(1,1,1);
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, gBGTexture);
		glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex2f(0,0);
		glTexCoord2f(1,0);
		glVertex2f(pixW,0);
		glTexCoord2f(1,1);
		glVertex2f(pixW,pixH);
		glTexCoord2f(0,1);
		glVertex2f(0,pixH);
		glEnd();
		glDisable(GL_TEXTURE_2D);
		PopOrthoPixels();
	}

	glLoadIdentity();
    glRotated(180.0*(-CameraAngleRadians/M_PI),0,0,1);
    glTranslatef(-CameraX,-CameraY,-CameraZ);

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
	/////////////////////////////////////////////////////////////////////////////////////////
	// Adjust the spring angles to help "lean" based on the x axis of the right thumbstick.
	LeanPerson(RightStick.x());

	glDisable(GL_BLEND);

	//////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////
	// yoga-specific rendering.

	// if we're in the pre-game screen...
	if (gCurPose == -1)
	{
		PushOrthoPixels();
		glColor3f(1,1,1);
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, gSplashTexture);
		glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex2f(0,0);
		glTexCoord2f(1,0);
		glVertex2f(pixW,0);
		glTexCoord2f(1,1);
		glVertex2f(pixW,pixH);
		glTexCoord2f(0,1);
		glVertex2f(0,pixH);
		glEnd();
		glDisable(GL_TEXTURE_2D);

		// if we're fading out, do that.
		if (gbFadingOut)
		{
			glEnable(GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			glColor4f(gFadeColor.x(), gFadeColor.y(), gFadeColor.z(), curFadeCompletion);
			glBegin(GL_QUADS);
			glVertex2f(0,0);
			glVertex2f(pixW,0);
			glVertex2f(pixW,pixH);
			glVertex2f(0,pixH);
			glEnd();
		}
		PopOrthoPixels();
	}

	// Draw the pose name
	if (gCurPose >= 0 && gCurPose < gPoses.size())
	{
		// iterate all the bodies, rendering them
		for (int i=0; i<objs.size(); i++)
		{
			if (objs[i]->name() != "floor")
			{
				glColor3f(1,1,1);
				Rigid_Body& obj = *objs[i];

				bool bDrawLines = true;
				if (gPoses[gCurPose].textures.find(obj.name()) != gPoses[gCurPose].textures.end())
				{
					GLuint tex = gPoses[gCurPose].textures[obj.name()].texture;
					glEnable(GL_TEXTURE_2D);
					glBindTexture(GL_TEXTURE_2D, tex);
					glEnable(GL_BLEND);
					glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
					bDrawLines = false;
				}
				else
				{
					glDisable(GL_TEXTURE_2D);
					glDisable(GL_BLEND);
				}
				glBody1v(obj.m_Pbody.get(), true);

				if (bDrawLines)
				{
					glColor3f(0,0,0);
					glBody1v(obj.m_Pbody.get(), false);
				}
			}
		}

		glDisable(GL_TEXTURE_2D);
		glDisable(GL_BLEND);

		PushOrthoPixels();

		// draw boxes for the UI to fit in.
		glLineWidth(3);
		glColor3f(0.5,0.5,0.5);
		glBegin(GL_LINE_LOOP);
		glVertex2f(kTopBoxLeft, pixH+kTopBoxTop);
		glVertex2f(kTopBoxRight, pixH+kTopBoxTop);
		glVertex2f(kTopBoxRight, pixH+kTopBoxBottom);
		glVertex2f(kTopBoxLeft, pixH+kTopBoxBottom);
		glEnd();
		glColor3f(1,1,1);
		glBegin(GL_QUADS);
		glVertex2f(kTopBoxRight, pixH+kTopBoxTop);
		glVertex2f(kTopBoxLeft, pixH+kTopBoxTop);
		glVertex2f(kTopBoxLeft, pixH+kTopBoxBottom);
		glVertex2f(kTopBoxRight, pixH+kTopBoxBottom);
		glEnd();
		glLineWidth(1);

		glLineWidth(3);
		glColor3f(0.5,0.5,0.5);
		glBegin(GL_LINE_LOOP);
		glVertex2f(pixW+kSideBoxLeft, pixH+kSideBoxTop);
		glVertex2f(pixW+kSideBoxRight, pixH+kSideBoxTop);
		glVertex2f(pixW+kSideBoxRight, pixH+kSideBoxBottom);
		glVertex2f(pixW+kSideBoxLeft, pixH+kSideBoxBottom);
		glEnd();
		glColor3f(1,1,1);
		glBegin(GL_QUADS);
		glVertex2f(pixW+kSideBoxRight, pixH+kSideBoxTop);
		glVertex2f(pixW+kSideBoxLeft, pixH+kSideBoxTop);
		glVertex2f(pixW+kSideBoxLeft, pixH+kSideBoxBottom);
		glVertex2f(pixW+kSideBoxRight, pixH+kSideBoxBottom);
		glEnd();
		glLineWidth(1);
		PopOrthoPixels();

		PushOrthoPixels();
		// draw the timer bar.
		if (gLevelTimeLeft > 0)
		{
			float curTimeBarRight = kTimeBarLeft + gLevelTimeLeft * (kTimeBarRight - kTimeBarLeft) / gLevelDuration;

			glColor3f(0.15f,0.f,0.75f);
			glBegin(GL_QUADS);
			glVertex2f(curTimeBarRight, pixH+kTimeBarTop);
			glVertex2f(kTimeBarLeft, pixH+kTimeBarTop);
			glVertex2f(kTimeBarLeft, pixH+kTimeBarBottom);
			glVertex2f(curTimeBarRight, pixH+kTimeBarBottom);
			glEnd();
		}

		glLineWidth(3);
		glColor3f(0,0,0);
		glBegin(GL_LINE_LOOP);
		glVertex2f(kTimeBarLeft-4, pixH+kTimeBarTop+4);
		glVertex2f(kTimeBarRight+4, pixH+kTimeBarTop+4);
		glVertex2f(kTimeBarRight+4, pixH+kTimeBarBottom-4);
		glVertex2f(kTimeBarLeft-4, pixH+kTimeBarBottom-4);
		glEnd();
		glLineWidth(1);

		if (gLevelTimeLeft > 0)
		{
			float curTimeBarRight = kTimeBarLeft + gLevelTimeLeft * (kTimeBarRight - kTimeBarLeft) / gLevelDuration;

			glColor3f(0.15f,0.f,0.75f);
			glBegin(GL_QUADS);
			glVertex2f(curTimeBarRight, pixH+kTimeBarTop);
			glVertex2f(kTimeBarLeft, pixH+kTimeBarTop);
			glVertex2f(kTimeBarLeft, pixH+kTimeBarBottom);
			glVertex2f(curTimeBarRight, pixH+kTimeBarBottom);
			glEnd();
		}

		// draw indications of what their controls are trying to do.
		DrawString(pixW+kMuscleTextLeft, pixH+kMuscleTextBottom0, "Muscle");
		DrawString(pixW+kMuscleTextLeft, pixH+kMuscleTextBottom1, "Tension");

		// draw the pose name.
		DrawString(kPoseNameLeft, pixH+kPoseNameBottom, (char*)gPoses[gCurPose].name.c_str());

		glLineWidth(3);
		glColor3f(0,0,0);
		glBegin(GL_LINE_LOOP);
		glVertex2f(pixW+kMuscleBarLeft-4, pixH+kMuscleBarTop+4);
		glVertex2f(pixW+kMuscleBarRight+4, pixH+kMuscleBarTop+4);
		glVertex2f(pixW+kMuscleBarRight+4, pixH+kMuscleBarBottom-4);
		glVertex2f(pixW+kMuscleBarLeft-4, pixH+kMuscleBarBottom-4);
		glEnd();
		glLineWidth(1);

		if (gLevelTimeLeft > 0)
		{
			float muscleTension = (0.5f * (RightStick.y() + 1.f));
			float curMuscleBarTop = kMuscleBarBottom + muscleTension * (kMuscleBarTop - kMuscleBarBottom);
			Vec3 color(muscleTension,0.f,1.f-muscleTension);

			glColor3fv(color.m_v);
			glBegin(GL_QUADS);
			glVertex2f(pixW+kMuscleBarRight, pixH+curMuscleBarTop);
			glVertex2f(pixW+kMuscleBarLeft, pixH+curMuscleBarTop);
			glVertex2f(pixW+kMuscleBarLeft, pixH+kMuscleBarBottom);
			glVertex2f(pixW+kMuscleBarRight, pixH+kMuscleBarBottom);
			glEnd();
		}

		// draw indications of lean.
		DrawString(pixW+kLeaningTextLeft, pixH+kLeaningTextBottom, "Lean");

		glLineWidth(3);
		glColor3f(0,0,0);
		glBegin(GL_LINE_LOOP);
		glVertex2f(pixW+kLeaningDialOuterLeft-7, pixH+kLeaningDialTop+4);
		glVertex2f(pixW+kLeaningDialOuterRight+7, pixH+kLeaningDialTop+4);
		glVertex2f(pixW+kLeaningDialInnerRight+4, pixH+kLeaningDialBottom-4);
		glVertex2f(pixW+kLeaningDialInnerLeft-4, pixH+kLeaningDialBottom-4);
		glEnd();
		glLineWidth(1);

		if (gLevelTimeLeft > 0)
		{
			float leaningTension = (0.5f * (RightStick.x() + 1.f));
			float curLeanOuterVal = kLeaningDialOuterLeft + leaningTension * (kLeaningDialOuterRight - kLeaningDialOuterLeft);
			float curLeanInnerVal = kLeaningDialInnerLeft + leaningTension * (kLeaningDialInnerRight - kLeaningDialInnerLeft);

			float leanIntense = fabs(RightStick.x());
			Vec3 color(leanIntense,0.f,1.f-leanIntense);

			glColor3fv(color.m_v);
			glLineWidth(4);
			glBegin(GL_LINES);
			glVertex2f(pixW+curLeanOuterVal, pixH+kLeaningDialTop);
			glVertex2f(pixW+curLeanInnerVal, pixH+kLeaningDialBottom);
			glEnd();
			glLineWidth(1);
		}

		// if we're fading out, do that.
		if (gbFadingOut)
		{
			glEnable(GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			glColor4f(gFadeColor.x(), gFadeColor.y(), gFadeColor.z(), curFadeCompletion);
			glBegin(GL_QUADS);
			glVertex2f(0,0);
			glVertex2f(pixW,0);
			glVertex2f(pixW,pixH);
			glVertex2f(0,pixH);
			glEnd();
		}

		PopOrthoPixels();
	}
	
	// if you're past the end of the poses, you won!
	else if (gCurPose == gPoses.size())
	{
		PushOrthoPixels();
		glColor3f(1,1,1);
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, gWinTexture);
		glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex2f(0,0);
		glTexCoord2f(1,0);
		glVertex2f(pixW,0);
		glTexCoord2f(1,1);
		glVertex2f(pixW,pixH);
		glTexCoord2f(0,1);
		glVertex2f(0,pixH);
		glEnd();
		glDisable(GL_TEXTURE_2D);
	}

	//////////////////////////////////////////////////////////////////////////////////////////////

    //DrawTimingBars();
    SwapBuffers(WindowDC);
}

/******************************** AppWndProc ********************************/
/* lame to expose this, but the alternatives all suck too                   */
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
    switch (msg)
    {
        // keyboard
        case WM_CHAR:
            break;

        case WM_MOUSEWHEEL:
        {
            float d = float((short)HIWORD(wParam))/float(WHEEL_DELTA);
            CameraZ *= 1.0f + d/10.0f;
            world_mouse = WorldFromWindowCoords(window_mouse);
            break;
        }
        case WM_LBUTTONDOWN:
        {
			SetCapture(hwnd);

            Vec2 body_space_position;
            Rigid_Body* rb = simulator.pick(body_space_position, world_mouse);
            if (rb)
            {
                spring = new Spring_Force(
                    Handle(world_body, world_mouse),
                    Handle(*rb, body_space_position), 
                    Globals.MouseSpringKs,
                    Globals.MouseSpringKd,
                    Globals.MouseSpringRestLength
                    );
                simulator.add(*spring);
            }
        } break;

        case WM_LBUTTONUP:
        {
			if(!(wParam & MK_MBUTTON & MK_RBUTTON))
			{
				ReleaseCapture();
			}

            ClearMouseSpring();
        } break;

        case WM_RBUTTONDOWN:
        {
			SetCapture(hwnd);

			if(spring)
            {
                forces.push_back(spring);
                spring = 0;
            }
            break;
        }
        case WM_MBUTTONDOWN:
        {
            MovingCamera = true;
        } break;
        case WM_MBUTTONUP:
        {
            MovingCamera = false;
        } break;

        case WM_RBUTTONUP:
        {
			if(!(wParam & MK_MBUTTON & MK_LBUTTON))
			{
				ReleaseCapture();
			}
        } break;
            
        case WM_KEYDOWN: {
            switch(wParam) {
                case VK_LEFT: {
                    CameraAngleRadians += float((5.0/180.0)*M_PI);
                    break;
                }
                case VK_RIGHT: {
                    CameraAngleRadians -= float((5.0/180.0)*M_PI);
                    break;
                }

                case 'M': {
                    // demo the audio stuff
                    if(0x8000 & GetKeyState(VK_SHIFT)) {
                        music.play("background");
                    } else {
                        zsoundPlay("content/sounds/bad.0002.wav");
                    }
                    break;
                }

				case 'N':
					LoadNextPose();
					break;

				case 'R':
					ReloadCurrentPose();
					break;

				case 'P':
					LoadPreviousPose();
					break;

                case VK_ESCAPE: {
                    AppQuit = true;
                    break;
                }
            }
            break;
        }
        case WM_SYSKEYDOWN:
        case WM_SYSKEYUP:
        case WM_KEYUP:
            break;

        case WM_MOUSEMOVE: {
			Vec2 new_window_mouse = Vec2(float(GET_X_LPARAM(lParam)),float(GET_Y_LPARAM(lParam)));
			if(MovingCamera)
			{
                Vec2 d_world = WorldFromWindowCoords(new_window_mouse) - world_mouse;
				CameraX -= d_world.x();
				CameraY -= d_world.y();
			}
            window_mouse = new_window_mouse;
            world_mouse = WorldFromWindowCoords(window_mouse);
            break;
        }

         
            /**************** below is stuff you probably don't care about ****************/
         
        case WM_DESTROY: {
            PostQuitMessage(0);
            break;
        }
        case WM_PAINT: {
            PAINTSTRUCT ps;
            BeginPaint(hwnd,&ps);
            Idle();
            EndPaint(hwnd,&ps);
            return 0L;
        }
        case WM_SIZE: {
            Resize(LOWORD(lParam),HIWORD(lParam));
            break;
        }
    }
    return (LONG)DefWindowProc(hwnd,msg,(LONG)wParam,(LONG)lParam);
}

/********************************** Resize **********************************/
/* called on WM_SIZE, and at the bottom of initialize                       */
void Resize( int w, int h )
{
    // compute our frustrum planes, the x,y plane values are on the near clip plane
    WindowWidth = w;
    WindowHeight = h;
    AspectRatio = float(w)/float(h);
    float ymax = ZNear * float(tan(FOVY/2.0));
    float ymin = -ymax;
    float xmin = AspectRatio * ymin;
    float xmax = AspectRatio * ymax;
   
    // load the viewport and projection matrix, gl stuff
    glViewport(0,0,w,h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // gluPerspective doesn't set the near plane as the projection
    // plane, so you can't do good math, so do it ourselves
    glFrustum(xmin,xmax,ymin,ymax,ZNear,ZFar);
    glMatrixMode(GL_MODELVIEW);
}

/******************************* Uninitialize *******************************/
/* called before the graphics is destroyed, but the window is already toast */
void Uninitialize( void )
{
}

/********************************* WinMain **********************************/

int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmdLine, int sw)
{
    // register and create our window
    WNDCLASS cls;
    if(!hPrev) {
        cls.hCursor        = LoadCursor(NULL,IDC_ARROW);
        cls.hIcon          = 0;
        cls.lpszMenuName   = 0;
        cls.lpszClassName  = AppName;
        cls.hbrBackground  = 0;
        cls.hInstance      = hInst;
        cls.style          = CS_VREDRAW | CS_HREDRAW | CS_DBLCLKS | CS_OWNDC;
        cls.lpfnWndProc    = (WNDPROC)AppWndProc;
        cls.cbWndExtra     = 0;
        cls.cbClsExtra     = 0;

        if (!RegisterClass(&cls))
            return FALSE;
    }

    DWORD Style = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
    RECT WindowRect = { 0, 0 };
    WindowRect.right = WindowWidth;
    WindowRect.bottom = WindowHeight;
    BOOL Menu = FALSE;

    AdjustWindowRect(&WindowRect,Style,Menu);
    
    WindowHandle = CreateWindow(AppName,AppName,Style,
                                0,50,
                                WindowRect.right-WindowRect.left,
                                WindowRect.bottom-WindowRect.top,
                                0,0,hInst,0);

    // create our gl stuff
    WindowDC = GetDC(WindowHandle);
    PIXELFORMATDESCRIPTOR FormatDescriptor = {
        sizeof (PIXELFORMATDESCRIPTOR), 1,
        PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
        PFD_TYPE_RGBA, 32,   // 32bpp
        0,0, 0,0, 0,0, 0,0,
        0, 0,0,0,0,
        24, 8,               // depth, stencil
        0,
        PFD_MAIN_PLANE,
        0,
        0,0,0
    };
    int FormatIndex = ChoosePixelFormat(WindowDC,&FormatDescriptor);
    SetPixelFormat(WindowDC,FormatIndex,&FormatDescriptor);
    WindowRC = wglCreateContext(WindowDC);
    wglMakeCurrent(WindowDC,WindowRC);

    // okay, ready to go
   
    Initialize();    // call the app
   
    ShowWindow(WindowHandle,sw);
   
    // pump the message queue
    MSG msg;
    while(!AppQuit) {
        if(PeekMessage(&msg,0,0,0,PM_REMOVE)) {
            if(msg.message == WM_QUIT)
                break;
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        } else {
            Idle();    // call the app
        }
    }

    Uninitialize();  // call the app
   
    // clean up gl stuff and exit
    wglMakeCurrent(0,0);
    wglDeleteContext(WindowRC);
    ReleaseDC(WindowHandle,WindowDC);
   
    return int(msg.wParam);
}

/********************** utilities that use our globals **********************/
Vec2 WorldFromWindowCoords( Vec2 window )
{
    float xn = float(window.x())/WindowWidth - 0.5f;
    float yn = float(WindowHeight-window.y())/WindowHeight - 0.5f;
    float xc = float(xn*2.0*CameraZ*tan(FOVY/2)*AspectRatio);
    float yc = float(yn*2.0*CameraZ*tan(FOVY/2));
    return Vec2(CameraX,CameraY) + Matrix22::rotation(CameraAngleRadians) * Vec2(xc,yc);
}
Vec2 WindowFromWorldCoords( Vec2 world )
{
    Vec2 c = Matrix22::rotation(-CameraAngleRadians) * (world - Vec2(CameraX,CameraY));
    float xn = float(c.x()/(2.0*CameraZ*tan(FOVY/2)*AspectRatio));
    float yn = float(c.y()/(2.0*CameraZ*tan(FOVY/2)));
    float xs = WindowWidth*(xn + 0.5f);
    float ys = WindowHeight - WindowHeight*(yn + 0.5f);
    return Vec2(xs,ys);
}
